﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace HeuristicLab.Problems.Instances.DataAnalysis.Regression.Matlab.Api.Types {
  public class MLDoubleArray : IMLValueVariable<double[,]> {
    private string[] _dataHeaders;

    #region Constructors

    public MLDoubleArray(string name, object values) {
      Name = name;
      if (!(values is double[,])) {
        throw new ArgumentException(string.Format("Invalid datatype for value: {0}", values.GetType()));
      }
      Data = values as double[,];

      _dataHeaders = new string[Data.GetLength(1)];
      for (int i = 0; i < _dataHeaders.Length; i++) {
        _dataHeaders[i] = string.Format("{0}_{1}", name, i);
      }
    }

    private MLDoubleArray(MLDoubleArray a1, MLDoubleArray a2) {
      Name = string.Format("{0};{1}", a1.Name, a2.Name);

      if (!ConcatIsPossible(a1, a2)) {
        throw new ArgumentException(string.Format("Cannot concat {0} and {1}. Different number of rows.", a1.Name, a2.Name));
      }

      var columns = a1.Columns + a2.Columns;
      _dataHeaders = new string[columns];
      for (int i = 0; i < columns; i++) {
        if (i < a1.Columns) {
          _dataHeaders[i] = a1.DataHeader[i];
        } else {
          _dataHeaders[i] = a2.DataHeader[i - a1.Columns];
        }
      }

      Data = new double[a1.Rows, columns];
      for (int i = 0; i < a1.Rows; i++) {
        for (int j = 0; j < columns; j++) {
          if (j < a1.Columns) {
            Data[i, j] = a1.Data[i, j];
          } else {
            Data[i, j] = a1.Data[i, j - a1.Columns];
          }
        }
      }
    }

    private MLDoubleArray(IEnumerable<MLDoubleArray> arrays) {
      StringBuilder sb = new StringBuilder();
      foreach (var array in arrays) {
        sb.Append(array.Name + ";");
      }
      Name = sb.ToString();

      var columns = 0;
      foreach (var array in arrays) {
        columns += array.Columns;
      }

      var dataHeaders = new List<string>();
      foreach (var array in arrays) {
        foreach (var header in array.DataHeader) {
          dataHeaders.Add(header);
        }
      }
      _dataHeaders = dataHeaders.ToArray();

      var rows = arrays.First().Rows;
      Data = new double[rows, columns];
      for (int i = 0; i < rows; i++) {
        var j = 0;
        foreach (var array in arrays) {
          for (int k = 0; k < array.Columns; k++) {
            Data[i, j] = array.Data[i, k];
            j++;
          }
        }
      }
    }
    #endregion


    /// <summary>
    /// Concats two given MLDoubleArrays if their number of rows equals.
    /// If throws an InvalidOperationException if they can't be concatenated
    /// </summary>
    /// <param name="a1"></param>
    /// <param name="a2"></param>
    /// <returns></returns>
    public static MLDoubleArray Concat(MLDoubleArray a1, MLDoubleArray a2) {
      if (ConcatIsPossible(a1, a2)) {
        throw new InvalidOperationException("Concatination is not possible because of different different number of their rows.");
      }
      return new MLDoubleArray(a1, a2);
    }

    public static MLDoubleArray Concat(IEnumerable<MLDoubleArray> arrays) {
      if (arrays.Count() < 1) {
        throw new ArgumentException("Given number of MLDoubleArrays is 0");
      }
      if (!ConcatIsPossible(arrays)) {
        throw new InvalidOperationException("Concatenation is not possible because of different different number of their rows.");
      }
      return new MLDoubleArray(arrays);
    }

    public static bool ConcatIsPossible(MLDoubleArray a1, MLDoubleArray a2) {
      return a1.Rows == a2.Rows;
    }

    public static bool ConcatIsPossible(IEnumerable<MLDoubleArray> arrays) {
      var first = arrays.First();
      foreach (var array in arrays) {
        if (!ConcatIsPossible(first, array)) {
          return false;
        }
      }
      return true;
    }


    #region Properties from the interface

    public string Name { get; set; }

    public double[,] Data { get; set; }

    public MLDatatype Datatype {
      get { return MLDatatype.DoubleArray; }
    }

    public string[] DataHeader {
      get { return _dataHeaders; }
    }
    #endregion

    public int Rows { get {
        return Data.GetLength(0);
      }
    }

    public int Columns {
      get {
        return Data.GetLength(1);
      }
    }
  }


}
